//Copyright (c) 2018 Ultimaker B.V.
//CuraEngine is released under the terms of the AGPLv3 or higher.

#ifndef UTILS_GETPATH_H
#define UTILS_GETPATH_H

#include <string>

namespace cura
{
std::string getPathName(const std::string& filePath);
}
#endif